import json
import math
import os
import pickle
import re
import warnings

import numpy as np
from sklearn.metrics import euclidean_distances

from gaussianlda.prior import Wishart
from scipy.linalg import solve_triangular
from scipy.special import gammaln
from sklearn.metrics.pairwise import cosine_similarity


class GaussianLDA:
    """
    Trained model.

    First train using the GaussianLDATrainer or GaussianLDAAliasTrainer.
    Then load using this class to get a GaussianLDA with the saved parameters for performing
    inference on new data without updating the parameters.

    """
    def __init__(self, vocab_embeddings, vocab, num_tables, alpha, kappa, table_counts, table_means, log_determinants,
                 table_cholesky_ltriangular_mat):
        # Vocab is used for outputting topics
        self.vocab = vocab

        # Dirichlet hyperparam
        self.alpha = alpha

        # dataVectors
        self.vocab_embeddings = vocab_embeddings
        self.embedding_size = vocab_embeddings.shape[1]
        # numIterations
        # K, num tables
        self.num_tables = num_tables
        # Number of customers observed at each table
        self.table_counts = table_counts
        # Mean vector associated with each table
        # This is the bayesian mean (i.e has the prior part too)
        self.table_means = table_means
        # log-determinant of covariance matrix for each table.
        # Since 0.5 * logDet is required in (see logMultivariateTDensity), that value is kept.
        self.log_determinants = log_determinants
        # Cholesky Lower Triangular Decomposition of covariance matrix associated with each table.
        self.table_cholesky_ltriangular_mat = table_cholesky_ltriangular_mat

        # Normal inverse wishart prior
        self.prior = Wishart(self.vocab_embeddings, kappa=kappa)

        # Cache k_0\mu_0\mu_0^T, only compute it once
        # Used in calculate_table_params()
        self.k0mu0mu0T = self.prior.kappa * np.outer(self.prior.mu, self.prior.mu)

        # Since we ignore the document's contributions to the global parameters when sampling,
        # we can precompute a whole load of parts of the likelihood calculation.
        # Table counts are not updated for the document in question, as it's assumed to make
        # a tiny contribution compared to the whole training corpus.
        k_n = self.prior.kappa + self.table_counts
        nu_n = self.prior.nu + self.table_counts
        self.scaleTdistrn = np.sqrt((k_n + 1.) / (k_n * (nu_n - self.embedding_size + 1.)))
        self.nu = self.prior.nu + self.table_counts - self.embedding_size + 1.
        # We can even scale the cholesky decomposition by scaleTdistrn
        self.scaled_table_cholesky_ltriangular_mat = \
            self.scaleTdistrn[:, np.newaxis, np.newaxis] * self.table_cholesky_ltriangular_mat

        self._topic_word_pdf_cache = {}

    @staticmethod
    def load_from_java(path, vocab_embeddings_path, vocab_path, alpha=None, kappa=None, nu=None, output_checks=False):
        """
        NB: This isn't fully working yet!
        I can't work out why not.

        Load a Gaussian LDA model trained and saved by the Gaussian LDA authors' original
        Java code, available at https://github.com/rajarshd/Gaussian_LDA/.
        The path given is to the directory containing all of the model files.

        Embeddings and vocab are not stored with the model, so need to be provided.
        They are given as paths to the files, stored in the same format expected by Gaussian
        LDA. This means you can use exactly the files you used to train the model.

        The files contain the output from each iteration of training. We just get the last iteration,
        or another one if explicitly requested.

        Unfortunately, the output does not store the hyperparameters alpha and kappa.
        kappa is needed to compute likelihoods under the topics and alpha is needed to perform
        topic inference, so you need to make sure these are correct to do correct inference.
        If not given, the default values from training are used.

        """
        # Load the vocab
        with open(vocab_path, "r") as f:
            vocab = [w.rstrip("\n") for w in f.readlines()]
        # Load the embeddings
        with open(vocab_embeddings_path, "r") as f:
            # Read the first line to check the dimensionality
            embedding_size = len(f.readline().split())
            # Put the embeddings into a Numpy array
            vocab_embeddings = np.zeros((len(vocab), embedding_size), dtype=np.float64)
            # We don't go back and read the first line, as it's never used by the model
            for i, line in enumerate(f):
                line = line.rstrip("\n")
                vocab_embeddings[i, :] = [float(v) for v in line.split()]
        #non_zero = np.where(np.sum(vocab_embeddings**2., axis=-1) > 0)
        #vocab_embeddings[non_zero] /= np.sqrt(np.sum(vocab_embeddings[non_zero]**2., axis=-1))[:, np.newaxis]

        if output_checks:
            # Sanity check the embeddings and vocab
            for w, top_word in enumerate(vocab[:3]):
                nns = np.argsort(-cosine_similarity(vocab_embeddings[w].reshape(1,-1), vocab_embeddings)[0])
                print("NNs to {}: {}".format(top_word, ", ".join(vocab[nb] for nb in nns[:5])))

        filenames = os.listdir(path)
        table_params_re = re.compile("\d+\.txt")
        table_params_filenames = [f for f in filenames if re.match(table_params_re, f)]
        table_params_filenames.sort()

        num_tables = len(table_params_filenames)

        if alpha is None:
            alpha = 1. / num_tables
        if kappa is None:
            kappa = 0.1
        if nu is None:
            nu = embedding_size

        # Create empty arrays to fill
        table_cholesky_ltriangular_mat = np.zeros((num_tables, embedding_size, embedding_size), dtype=np.float64)
        table_means = np.zeros((num_tables, embedding_size), dtype=np.float64)
        for table_filename in table_params_filenames:
            # Filename is of the form k.txt
            table_num = int(table_filename[:-4])
            with open(os.path.join(path, table_filename), "r") as f:
                lines = f.readlines()
            if len(lines) % (embedding_size+1) != 0:
                warnings.warn("Gaussian LDA model does not have the right number of lines "
                              "({} lines, expected {}+1)".format(len(lines), embedding_size))
            # The last D lines are the matrix and the one before is the mean
            # The first line contains the table mean
            table_mean = [float(v) for v in lines[0].rstrip("\n").split()]
            if len(table_mean) != embedding_size:
                raise ValueError("expected {}-size mean, but got {} for table {}".format(
                    embedding_size, len(table_mean), table_num))
            table_means[table_num, :] = table_mean
            # The remaining lines are the chol decomp of the cov matrix
            chol_mat = np.array([
                [float(v) for v in line.rstrip("\n").split()]
                for line in lines[1:]
            ], dtype=np.float64)
            table_cholesky_ltriangular_mat[table_num, :, :] = chol_mat

        if output_checks:
            for topic in range(num_tables):
                topic_nn = np.argsort(-cosine_similarity(table_means[topic].reshape(1, -1), vocab_embeddings)[0])[0]
                print("Topic {} centroid by cos sim: {}".format(topic, vocab[topic_nn]))

        # Load the total counts of customers at each table
        with open(os.path.join(path, "topic_counts.txt"), "r") as f:
            lines = f.readlines()
        # Should be K lines, with a count for each table
        if len(lines) != num_tables:
            raise IOError("expected {} lines in topic_counts.txt, got {}".format(num_tables, len(lines)))
        # The last K lines give us the final counts
        table_counts = np.array([float(v.rstrip("\n")) for v in lines], dtype=np.float64)

        # Compute the log determinants from the chol decomposition of the covariance matrices
        log_determinants = np.zeros(num_tables, dtype=np.float64)
        # Compute this in the same way as the Java code does
        for table in range(num_tables):
            # Log det of cov matrix is 2*log det of chol decomp
            k_n = float(table_counts[table]) + kappa
            nu_n = float(table_counts[table]) + nu
            scale_t_distrn = (k_n + 1.) / (k_n * (nu_n - embedding_size + 1.))
            log_determinants[table] = np.sum(np.log(np.diagonal(table_cholesky_ltriangular_mat[table, :, :]))) \
                                      + embedding_size * np.log(scale_t_distrn) / 2.

        # Initialize a model
        model = GaussianLDA(
            vocab_embeddings, vocab, num_tables, alpha, kappa,
            table_counts, table_means, log_determinants, table_cholesky_ltriangular_mat,
        )
        if output_checks:
            for table in range(num_tables):
                print("Computing probs for table {}".format(table))
                logprob = model.log_multivariate_tdensity(vocab_embeddings, table)
                top_word_ids = np.argsort(-logprob)
                top_words = [vocab[i] for i in top_word_ids]
                print("Top words for table {}: {}".format(table, top_words[:3]))
        return model

    @staticmethod
    def load(path):
        # Load JSON hyperparams
        with open(os.path.join(path, "params.json"), "r") as f:
            hyperparams = json.load(f)

        # Load numpy arrays for model parameters
        arrs = {}
        for name in [
            "table_counts", "table_means", "log_determinants", "table_cholesky_ltriangular_mat", "vocab_embeddings",
        ]:
            with open(os.path.join(path, "{}.pkl".format(name)), "rb") as f:
                arrs[name] = pickle.load(f)

        # Initialize a model
        model = GaussianLDA(
            arrs["vocab_embeddings"], hyperparams["vocab"], hyperparams["num_tables"], hyperparams["alpha"],
            hyperparams["kappa"],
            arrs["table_counts"], arrs["table_means"], arrs["log_determinants"],
            arrs["table_cholesky_ltriangular_mat"],
        )
        return model

    def sample(self, doc, num_iterations):
        """
        Run Gibbs sampler on a single document without updating global parameters.

        Input document should be a list of word strings.

        for num_iters:
            for each customer
                remove him from his old_table and update the table params.
                if old_table is empty:
                    remove table
                Calculate prior and likelihood for this customer sitting at each table
                sample for a table index
        """
        doc_ids = [self.vocab.index(w) for w in doc if w in self.vocab]
        table_assignments = list(np.random.randint(self.num_tables, size=len(doc_ids)))
        doc_table_counts = np.bincount(table_assignments, minlength=self.num_tables)

        for iteration in range(num_iterations):
            for w, cust_id in enumerate(doc_ids):
                # Remove custId from his old_table
                old_table_id = table_assignments[w]
                table_assignments[w] = -1  # Doesn't really make any difference, as only counts are used
                doc_table_counts[old_table_id] -= 1

                # Now calculate the prior and likelihood for the customer to sit in each table and sample
                # Go over each table
                counts = doc_table_counts[:] + self.alpha
                # Now calculate the likelihood for each table
                log_lls = self.log_multivariate_tdensity_tables(cust_id)
                # Add log prior in the posterior vector
                log_posteriors = np.log(counts) + log_lls
                # To prevent overflow, subtract by log(p_max).
                # This is because when we will be normalizing after exponentiating,
                # each entry will be exp(log p_i - log p_max )/\Sigma_i exp(log p_i - log p_max)
                # the log p_max cancels put and prevents overflow in the exponentiating phase.
                posterior = np.exp(log_posteriors - log_posteriors.max())
                posterior /= posterior.sum()
                # Now sample an index from this posterior vector.
                new_table_id = np.random.choice(self.num_tables, p=posterior)

                # Now have a new assignment: add its counts
                doc_table_counts[new_table_id] += 1
                table_assignments[w] = new_table_id
        return table_assignments

    def log_multivariate_tdensity(self, x, table_id):
        """
        Gaussian likelihood for a table-embedding pair when using Cholesky decomposition.

        """
        if x.ndim > 1:
            logprobs = np.zeros(x.shape[0], dtype=np.float64)
            for i in range(x.shape[0]):
                logprobs[i] = self.log_multivariate_tdensity(x[i], table_id)
            return logprobs

        nu = self.nu[table_id]
        # first calculate (x-mu)
        x_minus_mu = x - self.table_means[table_id]
        ltriangular_chol = self.scaled_table_cholesky_ltriangular_mat[table_id]
        solved = solve_triangular(ltriangular_chol, x_minus_mu, check_finite=False)
        # Now take xTx (dot product)
        val = (solved ** 2.).sum(-1)

        logprob = gammaln((nu + self.embedding_size) / 2.) - \
                  (
                          gammaln(nu / 2.) +
                          self.embedding_size / 2. * (np.log(nu) + np.log(math.pi)) +
                          self.log_determinants[table_id] +
                          (nu + self.embedding_size) / 2. * np.log(1. + val / nu)
                  )
        return logprob

    def log_multivariate_tdensity_tables(self, x):
        """
        Gaussian likelihood for a table-embedding pair when using Cholesky decomposition.
        This version computes the likelihood for all tables in parallel.

        If x is an int, it is treated as an index to the vocabulary of known embeddings. This
        will often be a more efficient way to call this repeated times, since it allows us to
        caches the concept-word probs for words in the vocabulary. This is possible, since
        the concept parameters are never updated.

        """
        if type(x) is int:
            # Treat as vocab index and cache the probability
            if x not in self._topic_word_pdf_cache:
                # Not already computed this: compute now and cache the logprobs
                self._topic_word_pdf_cache[x] = self.log_multivariate_tdensity_tables(self.vocab_embeddings[x])
            return self._topic_word_pdf_cache[x]

        # Since I am storing lower triangular matrices, it is easy to calculate (x-\mu)^T\Sigma^-1(x-\mu)
        # therefore I am gonna use triangular solver first calculate (x-mu)
        x_minus_mu = x[None, :] - self.table_means
        # We can't do solve_triangular for all matrices at once in scipy
        val = np.zeros(self.num_tables, dtype=np.float64)
        for table in range(self.num_tables):
            table_solved = solve_triangular(self.scaled_table_cholesky_ltriangular_mat[table], x_minus_mu[table])
            # Now take xTx (dot product)
            val[table] = (table_solved ** 2.).sum()

        logprob = gammaln((self.nu + self.embedding_size) / 2.) - \
                  (
                          gammaln(self.nu / 2.) +
                          self.embedding_size / 2. * (np.log(self.nu) + np.log(math.pi)) +
                          self.log_determinants +
                          (self.nu + self.embedding_size) / 2. * np.log(1. + val / self.nu)
                  )
        return logprob



import numpy as np


def generate_table(w, num_tables):
    ## The implementation copied from Java GaussianLDA
    # Generate a new prob array and alias table
    alias = np.zeros(num_tables, dtype=np.int32)
    # Contains proportions and alias probabilities
    prob = np.zeros(num_tables, dtype=np.float32)

    # 1. Create two worklists, Small and Large
    # 2. Multiply each probability by n
    p = (w * num_tables) / w.sum()
    # 3. For each scaled probability pi:
    #   a. If pi<1, add i to Small.
    #   b. Otherwise(pi>=1), add i to Large.
    small, large = [], []
    for i, pi in enumerate(p):
        if pi < 1.:
            small.append(i)
        else:
            large.append(i)
    #small = list(np.where(p < 1.)[0])
    #large = list(np.where(p >= 1.)[0])
    # 4. While Small and Large are not empty : (Large might be emptied first)
    #    a. Remove the first element from Small; call it l.
    #    b. Remove the first element from Large; call it g.
    #    c. Set Prob[l] = pl.
    #    d. Set Alias[l] = g.
    #    e. Set pg : = (pg + pl)−1. (This is a more numerically stable option.)
    #    f. If pg<1, add g to Small.
    #    g. Otherwise(pg≥1), add g to Large.
    while len(small) and len(large):
        l = small.pop(0)
        g = large.pop(0)
        prob[l] = p[l]
        alias[l] = g
        p[g] = (p[g] + p[l]) - 1.
        if p[g] < 1.:
            small.append(g)
        else:
            large.append(g)

    # 5. While Large is not empty :
    #    a. Remove the first element from Large; call it g.
    #    b. Set Prob[g] = 1.
    if len(large):
        prob[large] = 1.

    # 6. While Small is not empty : This is only possible due to numerical instability.
    #    a. Remove the first element from Small; call it l.
    #    b. Set Prob[l] = 1.
    if len(small):
        prob[small] = 1.

    return alias, prob


if __name__ == "__main__":
    from .chol_alias import VoseAlias

    alias = VoseAlias.create(5)

    # Use a predefined set of probabilities
    known_probs = np.array([0.1, 0.1, 0.4, 0.05, 0.35])
    print("Probabilities: ", known_probs)

    # Compute the tables in the same way that the updater does
    log_likelihoods = np.log(known_probs)
    ll_max = log_likelihoods.max()
    log_likelihoods_scaled = log_likelihoods - ll_max
    # Exp them all to get the new w array
    w = np.exp(log_likelihoods_scaled)
    # Sum of likelihood has to be scaled back to its original sum, before we subtracted the max log
    # likelihood_sum = np.exp(np.log(w.sum()) + ll_max)
    # NB Not doing this now, but following what the Java impl does, however odd that seems
    likelihood_sum = w.sum()
    # Update the alias and probs using this w
    new_alias, new_prob = generate_table(w, 5)
    # Update the parameters of this word's alias
    np.copyto(alias.log_likelihoods.np, log_likelihoods)
    np.copyto(alias.alias.np, new_alias)
    np.copyto(alias.prob.np, new_prob)
    np.copyto(alias.likelihood_sum.np, likelihood_sum)

    # Draw lots of samples from the alias
    print("Drawing 1000 samples")
    sample_counts = np.zeros(5, dtype=np.float32)
    for sample_num in range(1000):
        sample_counts[alias.sample_vose()] += 1
    print("Samples:", sample_counts)
    print("Proportions:", sample_counts/sample_counts.sum())
    # These should be close to the probabilities we put in
    # If so, the sampler is working
    # For comparison, use numpy to sample
    print("Sampling using Numpy")
    sample_counts = np.zeros(5, dtype=np.float32)
    for sample_num in range(1000):
        sample_counts[np.random.choice(5, p=known_probs)] += 1
    print("Samples:", sample_counts)
    print("Proportions:", sample_counts/sample_counts.sum())




import logging
import progressbar as pb

import numpy as np
from numpy.random import default_rng

from choldate import cholupdate, choldowndate


def get_logger(name):
    """
    Convenience function to make it easier to create new loggers.

    :param name: logging system logger name
    :return:
    """
    level = logging.INFO

    # Prepare a logger
    log = logging.getLogger(name)
    log.setLevel(level)

    fmt = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'

    # If coloredlogs is available, enable it
    # We don't make this one of Pimlico's core dependencies, but simply allow
    # it to be installed manually on the system and use it if it's there
    try:
        import coloredlogs
    except ImportError:
        # No coloredogs: never mind
        # Check there's a handler and add a stream handler if not
        if not log.handlers:
            # Just log to the console
            sh = logging.StreamHandler()
            sh.setLevel(logging.DEBUG)
            # Put a timestamp on everything
            formatter = logging.Formatter(fmt)
            sh.setFormatter(formatter)

            log.addHandler(sh)
    else:
        coloredlogs.install(
            level=level, log=log, fmt=fmt
        )

    return log


def get_progress_bar(maxval, title=None, counter=False, show_progress=True):
    # Provides an easy way to switch to showing no progress bar
    if not show_progress:
        return lambda x: x

    widgets = []
    if title is not None:
        widgets.append("%s: " % title)
    if maxval is not pb.UnknownLength:
        widgets.extend([pb.Percentage(), ' ', pb.Bar(marker=pb.RotatingMarker())])
    if counter:
        widgets.extend([' (', pb.Counter(), ')'])
    if maxval is not pb.UnknownLength:
        widgets.extend([' ', pb.ETA()])
    pbar = pb.ProgressBar(widgets=widgets, maxval=maxval)
    return pbar


def chol_rank1_update(L, x):
    # choldate only uses float64
    cholupdate(L.T, x.copy())


def chol_rank1_downdate(L, x):
    choldowndate(L.T, x.copy())


def sum_logprobs(logprobs):
    """Sum probabilities represented as logprobs.

    Avoids underflow by a standard trick. Not massively efficient: I think there's a
    more efficient way to do this.

    Returns the log of the summed probabilities.

    """
    max_prob = logprobs.max()
    return np.log(np.sum(np.exp(logprobs - max_prob))) + max_prob


class BatchedRands:
    """
    Draw multiple random samples from Numpy at once to speed up repeated
    random sampling.

    Change how many are drawn at once by setting batch_size. Very small
    numbers (<10) will be very slow, like just drawing repeated samples. Numbers
    more like 100 will be better. 1000 probably gives a small speedup, but beyond
    that there's not much difference.

    """
    def __init__(self, batch_size=1000):
        self.batch_size = batch_size
        self.rng = default_rng()
        self._it = iter(self)

    def __iter__(self):
        while True:
            batch = self.rng.random(self.batch_size)
            for v in batch:
                yield v

    def random(self):
        return next(self._it)

    def integer(self, high):
        return int(high * self.random())

    def choice(self, p):
        rand = self.random()
        return np.argmax(np.cumsum(p) > rand)

    def choice_cum(self, p):
        """
        Like choice, but probabilities are given in cumulative form. This is useful
        if you need to draw several times from the same dist, to avoid recomputing
        the cumsum. For a large number of samples, this can give a largeish speedup
        (like 2x).

        """
        rand = self.random()
        return np.argmax(p > rand)


class BatchedRandInts:
    """
    Similar to BatchedRands, but generates arrays of integers. The size of the
    arrays varies, but the range of the integers is fixed.

    """
    def __init__(self, high, batch_size=1000):
        self.high = high
        self.batch_size = batch_size
        self.rng = default_rng()
        self._new_batch()

    def _new_batch(self):
        self._current_id = 0
        self._batch = self.rng.integers(self.high, size=self.batch_size)

    def integers(self, size):
        self._current_id += size
        if self._current_id > self.batch_size:
            # Reached end of batch in middle of range
            self._new_batch()
            self._current_id += size

        return self._batch[self._current_id-size:self._current_id]



import numpy as np
from scipy.linalg import solve_triangular


def corpus_categorical_mean_ll(corpus, table_assignments, table_word_logprobs, check_probs=True):
    """
    Compute the perplexity of a Gaussian LDA model on a given corpus, together with the
    assignment of topics to each word.

    This is computed just from the probability of the words given the assignments, not
    taking into account the priors on the assignments or the Gaussians. In this, we
    follow Util.calculateAvgLL() in the Java implementation.

    The perplexity is exp(-X).

    For comparability with other models, we treat the
    Gaussians as if they were categorical distributions over the vocabulary of known
    words in the corpus. This means that the word likelihoods sum to one over the vocabulary,
    which is not the case if you take the point PDF density from the Gaussian (as the
    original Gaussian LDA Java code does).

    This means that this perplexity is comparable, for example, with a similar value for
    standard LDA over the same corpus and the same vocabulary.

    """
    if check_probs:
        # Check that the distribution sums (roughly) to 1
        for table_logprobs in table_word_logprobs:
            max_logprob = table_logprobs.max()
            table_logprobs -= max_logprob
            table_total_scaled_prob = np.sum(np.exp(table_logprobs))
            total_prob = np.exp(np.log(table_total_scaled_prob) + max_logprob)
            # This should be roughly 1
            if abs(total_prob - 1.) > 1e-5:
                raise ValueError("Table-word distribution probs should sum to 1. Note that they're given as logprobs")

    total_logprob = 0.
    num_words = 0
    for doc, doc_tables in zip(corpus, table_assignments):
        for word_id, table_id in zip(doc, doc_tables):
            total_logprob += table_word_logprobs[table_id, word_id]
            num_words += 1

    return total_logprob / num_words


def calculate_avg_ll(corpus, table_assignments, embeddings, table_means, table_cholesky_ltriangular_mat, prior, table_counts_per_doc):
    """
    Calculates corpus perplexity (avg. log-likelihood)

    Reproduction of Gaussian_LDA's Util.calculateAvgLL(). It's not really perplexity, but
    the mean LL of the words given the sampled topics.

    """
    num_tables = table_means.shape[0]
    embedding_size = embeddings.shape[1]
    # Sum up the total number of customers per table
    n_k = table_counts_per_doc.sum(axis=1)

    scalar = n_k + prior.nu - embedding_size
    # now divide the choleskies by sqrt(scalar)
    scaled_choleskies = table_cholesky_ltriangular_mat / np.sqrt(scalar)[:, np.newaxis, np.newaxis]

    # logDensity of mulitvariate normal is given by -0.5*(log D + K*log(2*\pi)+(x-\mu)^T\Sigma^-1(x-\mu))
    # calculate log D for all table from cholesky
    log_det = np.zeros(num_tables, dtype=np.float64)
    for table in range(num_tables):
        log_det[table] = np.sum(np.log(np.diagonal(scaled_choleskies[table])))

    # Cache the table-word pairs' log densities to speed this up
    log_density_cache = {}

    total_log_ll = 0.
    total_words = 0
    for doc, tables in zip(corpus, table_assignments):
        for word, table in zip(doc, tables):
            if (table, word) in log_density_cache:
                log_density = log_density_cache[(table, word)]
            else:
                # Do exactly what the Java code does
                x = embeddings[word]
                x_minus_mu = x - table_means[table]
                ltriangular_chol = scaled_choleskies[table]
                solved = solve_triangular(ltriangular_chol, x_minus_mu, check_finite=False, lower=True)
                val = np.sum(solved ** 2.)
                log_density = 0.5 * (val + embedding_size * np.log(2. * np.pi)) + log_det[table]

                log_density_cache[(table, word)] = log_density

            total_log_ll -= log_density
            total_words += 1

    return total_log_ll / total_words
